export function createController(options) {
    return new PreviewIframeControllerImpl(options);
}
class PreviewIframeControllerImpl {
    constructor(options) {
        this.options = options;
        this.previewBootstrapped = false;
        this.waitingForBootstrapped = false;
        this.lastMessage = null;
        this.onWindowMessage = (event) => {
            const data = event.data;
            const { listener } = this.options;
            switch (data.kind) {
                case "bootstrapped":
                    this.onBootstrapped();
                    break;
                case "before-render":
                case "action":
                case "log-message":
                    listener(data);
                    break;
                case "renderer-updated":
                    this.clearExpectRenderTimeout();
                    listener({
                        kind: "update",
                        rendering: {
                            kind: "success",
                            variantKey: data.variantKey,
                            variants: data.variants,
                        },
                    });
                    break;
                case "rendering-error":
                    this.clearExpectRenderTimeout();
                    listener({
                        kind: "log-message",
                        level: "error",
                        timestamp: Date.now(),
                        message: data.message,
                    });
                    break;
                case "vite-error":
                    listener({
                        kind: "log-message",
                        level: "error",
                        timestamp: Date.now(),
                        message: generateMessageFromViteError(data.payload.err),
                    });
                    break;
                case "vite-before-update":
                    listener({
                        kind: "update",
                        rendering: null,
                    });
                    break;
            }
        };
    }
    start() {
        window.addEventListener("message", this.onWindowMessage);
        this.resetIframe();
    }
    stop() {
        window.removeEventListener("message", this.onWindowMessage);
    }
    showLoading() {
        this.send({
            kind: "show-loading",
        });
    }
    loadComponent(options) {
        this.send({
            kind: "render",
            ...options,
        });
    }
    send(message) {
        var _a;
        this.lastMessage = message;
        if (!this.previewBootstrapped && !this.waitingForBootstrapped) {
            this.resetIframe();
            return;
        }
        const iframeWindow = (_a = this.options.getIframe()) === null || _a === void 0 ? void 0 : _a.contentWindow;
        if (!iframeWindow) {
            return;
        }
        iframeWindow.postMessage(message, document.location.href);
        if (message.kind === "render") {
            this.clearExpectRenderTimeout();
            this.expectRenderTimeout = setTimeout(() => {
                console.warn("Expected render did not occur after 5 seconds. Reloading iframe...");
                this.resetIframe();
            }, 5000);
        }
    }
    resetIframe() {
        const iframe = this.options.getIframe();
        this.previewBootstrapped = false;
        if (!iframe) {
            return;
        }
        this.waitingForBootstrapped = true;
        iframe.src = `/preview/?t=${Date.now()}`;
    }
    onBootstrapped() {
        this.previewBootstrapped = true;
        this.waitingForBootstrapped = false;
        if (this.lastMessage) {
            this.send(this.lastMessage);
        }
        this.options.listener({
            kind: "bootstrapped",
        });
    }
    clearExpectRenderTimeout() {
        if (this.expectRenderTimeout) {
            clearTimeout(this.expectRenderTimeout);
            this.expectRenderTimeout = null;
        }
    }
}
function generateMessageFromViteError(err) {
    let message = err.message + (err.stack ? `\n${err.stack}` : "");
    // Remove any redundant line breaks (but not spaces,
    // which could be useful indentation).
    message = message.replace(/^\n+/g, "\n").trim();
    const stripPrefix = "Internal server error: ";
    if (message.startsWith(stripPrefix)) {
        message = message.substr(stripPrefix.length);
    }
    if (/^Transform failed with \d+ errors?:?\n.*/.test(message)) {
        const lineBreakPosition = message.indexOf("\n");
        message = message.substring(lineBreakPosition + 1);
    }
    const lineBreakPosition = message.indexOf("\n");
    let title;
    let rest;
    if (lineBreakPosition > -1) {
        title = message.substr(0, lineBreakPosition).trim();
        rest = message.substr(lineBreakPosition + 1);
    }
    else {
        title = message;
        rest = "";
    }
    if (title.endsWith(":") || title.endsWith(".")) {
        title = title.substr(0, title.length - 1).trim();
    }
    // Note: this isn't relevant to all browsers.
    if (rest.startsWith(`Error: ${title}\n`)) {
        rest = rest.substr(rest.indexOf("\n") + 1);
    }
    return `${title}${rest ? `\n\n${rest}` : ""}`;
}
