"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.findWorkspaceRoot = exports.createWorkspace = exports.loadPreviewEnv = exports.generateComponentId = void 0;
const api_1 = require("@previewjs/api");
const type_analyzer_1 = require("@previewjs/type-analyzer");
const cookie_parser_1 = __importDefault(require("cookie-parser"));
const express_1 = __importDefault(require("express"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const get_port_1 = __importDefault(require("get-port"));
const path_1 = __importDefault(require("path"));
const analyze_project_1 = require("./analyze-project");
const compute_props_1 = require("./compute-props");
const persisted_state_1 = require("./persisted-state");
const previewer_1 = require("./previewer");
const router_1 = require("./router");
var component_id_1 = require("./component-id");
Object.defineProperty(exports, "generateComponentId", { enumerable: true, get: function () { return component_id_1.generateComponentId; } });
var preview_env_1 = require("./preview-env");
Object.defineProperty(exports, "loadPreviewEnv", { enumerable: true, get: function () { return preview_env_1.loadPreviewEnv; } });
async function createWorkspace({ versionCode, rootDirPath, reader, frameworkPlugin, logLevel, middlewares, onReady, persistedStateManager = new persisted_state_1.LocalFilePersistedStateManager(), }) {
    const expectedPluginApiVersion = 3;
    if (!frameworkPlugin.pluginApiVersion ||
        frameworkPlugin.pluginApiVersion < expectedPluginApiVersion) {
        throw new Error(`Preview.js framework plugin ${frameworkPlugin.name} is incompatible with this version of Preview.js. Please upgrade it.`);
    }
    else if (frameworkPlugin.pluginApiVersion > expectedPluginApiVersion) {
        throw new Error(`Preview.js framework plugin ${frameworkPlugin.name} is too recent. Please upgrade Preview.js or use an older version of ${frameworkPlugin.name}.`);
    }
    if (frameworkPlugin.transformReader) {
        reader = frameworkPlugin.transformReader(reader, rootDirPath);
    }
    const collected = {};
    const typeAnalyzer = (0, type_analyzer_1.createTypeAnalyzer)({
        reader,
        rootDirPath,
        collected,
        specialTypes: frameworkPlugin.specialTypes,
        tsCompilerOptions: frameworkPlugin.tsCompilerOptions,
    });
    const router = new router_1.ApiRouter();
    router.onRequest(api_1.localEndpoints.GetInfo, async () => {
        const separatorPosition = versionCode.indexOf("-");
        if (separatorPosition === -1) {
            throw new Error(`Unsupported version code format: ${versionCode}`);
        }
        const platform = versionCode.substr(0, separatorPosition);
        const version = versionCode.substr(separatorPosition + 1);
        return {
            appInfo: {
                platform,
                version,
            },
        };
    });
    router.onRequest(api_1.localEndpoints.GetState, persistedStateManager.get);
    router.onRequest(api_1.localEndpoints.UpdateState, persistedStateManager.update);
    router.onRequest(api_1.localEndpoints.ComputeProps, async ({ filePath, componentName }) => {
        const component = (await frameworkPlugin.detectComponents(typeAnalyzer, [
            path_1.default.join(rootDirPath, filePath),
        ])).find((c) => c.name === componentName);
        if (!component) {
            return {
                types: {
                    props: type_analyzer_1.UNKNOWN_TYPE,
                    all: {},
                },
                args: [],
            };
        }
        return (0, compute_props_1.computeProps)({
            component,
        });
    });
    const previewer = new previewer_1.Previewer({
        reader,
        rootDirPath,
        previewDirPath: path_1.default.join(__dirname, "..", "..", "iframe", "preview"),
        frameworkPlugin,
        logLevel,
        middlewares: [
            express_1.default.json(),
            (0, cookie_parser_1.default)(),
            express_1.default
                .Router()
                .use("/monaco-editor", express_1.default.static(path_1.default.join(__dirname, "..", "monaco-editor"))),
            async (req, res, next) => {
                if (req.path.startsWith("/api/")) {
                    res.json(await router.handle(req.path.substr(5), req.body, req, res));
                }
                else {
                    next();
                }
            },
            ...middlewares,
        ],
        onFileChanged: (absoluteFilePath) => {
            const filePath = path_1.default.relative(rootDirPath, absoluteFilePath);
            for (const name of Object.keys(collected)) {
                if (name.startsWith(`${filePath}:`)) {
                    delete collected[name];
                }
            }
        },
    });
    const workspace = {
        rootDirPath,
        reader,
        typeAnalyzer,
        frameworkPlugin,
        preview: {
            start: async (allocatePort) => {
                const port = await previewer.start(async () => {
                    const port = allocatePort ? await allocatePort() : 0;
                    return (port ||
                        (await (0, get_port_1.default)({
                            port: get_port_1.default.makeRange(3140, 4000),
                        })));
                });
                return {
                    url: () => `http://localhost:${port}`,
                    stop: async (options) => {
                        await previewer.stop(options);
                    },
                };
            },
        },
        components: {
            list: (options) => {
                return (0, analyze_project_1.analyzeProject)(workspace, options);
            },
        },
        dispose: async () => {
            typeAnalyzer.dispose();
        },
    };
    if (onReady) {
        await onReady({
            router,
            workspace,
        });
    }
    return workspace;
}
exports.createWorkspace = createWorkspace;
/**
 * Returns the absolute directory path of the closest ancestor containing node_modules.
 */
function findWorkspaceRoot(absoluteFilePath) {
    let dirPath = path_1.default.resolve(absoluteFilePath);
    while (dirPath !== path_1.default.dirname(dirPath)) {
        if (fs_extra_1.default.existsSync(path_1.default.join(dirPath, "package.json"))) {
            return dirPath;
        }
        dirPath = path_1.default.dirname(dirPath);
    }
    return null;
}
exports.findWorkspaceRoot = findWorkspaceRoot;
