"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.loadPreviewEnv = void 0;
const config_1 = require("@previewjs/config");
const fs_extra_1 = __importDefault(require("fs-extra"));
const path_1 = __importDefault(require("path"));
async function loadPreviewEnv({ rootDirPath, setupEnvironment, frameworkPluginFactories, }) {
    const previewEnv = await setupEnvironment({ rootDirPath });
    let frameworkPlugin = await (0, config_1.readConfig)(rootDirPath).frameworkPlugin;
    fallbackToDefault: if (!frameworkPlugin) {
        const dependencies = await extractPackageDependencies(rootDirPath);
        for (const candidate of frameworkPluginFactories || []) {
            if (await candidate.isCompatible(dependencies)) {
                frameworkPlugin = await candidate.create();
                break fallbackToDefault;
            }
        }
        return null;
    }
    return {
        previewEnv,
        frameworkPlugin,
    };
}
exports.loadPreviewEnv = loadPreviewEnv;
async function extractPackageDependencies(rootDirPath) {
    const packageJsonPath = path_1.default.join(rootDirPath, "package.json");
    if (!(await fs_extra_1.default.pathExists(packageJsonPath))) {
        return {};
    }
    let { dependencies, devDependencies, peerDependencies } = JSON.parse(await fs_extra_1.default.readFile(packageJsonPath, "utf8"));
    const allDependencies = {
        ...dependencies,
        ...devDependencies,
        ...peerDependencies,
    };
    return Object.fromEntries(Object.entries(allDependencies).map(([name, version]) => {
        let majorVersion;
        if (typeof version !== "string") {
            majorVersion = 0;
        }
        else if (version.startsWith("^") || version.startsWith("~")) {
            majorVersion = parseInt(version.slice(1));
        }
        else {
            majorVersion = parseInt(version);
        }
        const readInstalledVersion = async () => {
            try {
                const moduleEntryPath = require.resolve(name, {
                    paths: [rootDirPath],
                });
                let packagePath = moduleEntryPath;
                let packageJsonPath = null;
                while (packagePath !== path_1.default.dirname(packagePath)) {
                    const candidatePackageJsonPath = path_1.default.join(packagePath, "package.json");
                    if (fs_extra_1.default.existsSync(candidatePackageJsonPath)) {
                        packageJsonPath = candidatePackageJsonPath;
                        break;
                    }
                    packagePath = path_1.default.dirname(packagePath);
                }
                if (!packageJsonPath) {
                    throw new Error(`No package.json path found from: ${moduleEntryPath}`);
                }
                const packageInfo = JSON.parse(await fs_extra_1.default.readFile(packageJsonPath, "utf8"));
                const version = packageInfo["version"];
                if (!version || typeof version !== "string") {
                    throw new Error(`Invalid version found for package: ${packageJsonPath}`);
                }
                return version;
            }
            catch (e) {
                console.error(`Unable to read installed version of package: ${name}`, e);
                return null;
            }
        };
        return [name, { majorVersion, readInstalledVersion }];
    }));
}
