"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.virtualPlugin = void 0;
const fs_extra_1 = __importDefault(require("fs-extra"));
const path_1 = __importDefault(require("path"));
const vite_1 = require("vite");
const VIRTUAL_PREFIX = `/@previewjs-virtual:`;
const jsExtensions = new Set([".js", ".jsx", ".ts", ".tsx"]);
function virtualPlugin(options) {
    const { reader, rootDirPath } = options;
    return {
        name: "previewjs:virtual-fs",
        resolveId: async function (id, importer) {
            const virtualImporter = (importer === null || importer === void 0 ? void 0 : importer.startsWith(VIRTUAL_PREFIX)) || false;
            if (id.indexOf(`/node_modules/`) !== -1) {
                return null;
            }
            if (id.startsWith(VIRTUAL_PREFIX)) {
                id = id.slice(VIRTUAL_PREFIX.length);
            }
            const extension = path_1.default.extname(id);
            let absoluteId;
            if (path_1.default.isAbsolute(id)) {
                absoluteId = id;
            }
            else {
                if (!importer) {
                    return null;
                }
                if (virtualImporter) {
                    importer = importer.slice(VIRTUAL_PREFIX.length);
                }
                if (extension && !jsExtensions.has(extension) && extension !== ".svg") {
                    // Virtual files mess with CSS processors like postcss.
                    return path_1.default.join(path_1.default.dirname(importer), id);
                }
                absoluteId = path_1.default.join(path_1.default.dirname(importer), id);
            }
            const resolved = await resolveAbsoluteModuleId(absoluteId);
            if (resolved) {
                const [absoluteFilePath] = resolved;
                if ((await fs_extra_1.default.pathExists(absoluteFilePath)) && !virtualImporter) {
                    // This file doesn't need to be virtual.
                    return null;
                }
                if (!absoluteId.endsWith(extension)) {
                    absoluteId += extension;
                }
                const resolvedId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
                return resolvedId;
            }
            return null;
        },
        load: async function (id) {
            if (id.indexOf(`/node_modules/`) !== -1) {
                return null;
            }
            if (id.startsWith(VIRTUAL_PREFIX)) {
                id = id.slice(VIRTUAL_PREFIX.length);
            }
            const resolved = await resolveAbsoluteModuleId(id);
            if (!resolved) {
                // This could be a file handled by another plugin, e.g. CSS modules.
                return null;
            }
            const [absoluteFilePath, entry] = resolved;
            if (entry.kind !== "file") {
                console.error(`Unable to read from ${absoluteFilePath}`);
                return null;
            }
            let source = await entry.read();
            const fileExtension = path_1.default.extname(absoluteFilePath);
            if (!jsExtensions.has(fileExtension)) {
                return source;
            }
            const moduleExtension = path_1.default.extname(id);
            return {
                code: 
                // We run an esbuild transform for .js or no extension
                // because it may include JSX. Otherwise, let plugins
                // decide whether to use esbuild or not.
                moduleExtension === "" || moduleExtension === ".js"
                    ? (await (0, vite_1.transformWithEsbuild)(source, absoluteFilePath, {
                        loader: "tsx",
                        format: "esm",
                        sourcefile: path_1.default.relative(rootDirPath, absoluteFilePath),
                        ...options.esbuildOptions,
                    })).code
                    : source,
            };
        },
        handleHotUpdate: async function (context) {
            const moduleGraph = options.moduleGraph();
            if (!moduleGraph) {
                return;
            }
            const filePath = path_1.default.relative(rootDirPath, context.file);
            const absoluteFilePath = path_1.default.join(rootDirPath, filePath);
            const entry = await reader.read(absoluteFilePath);
            if (!entry || entry.kind !== "file") {
                return;
            }
            // Note: backslash handling is Windows-specific.
            const virtualModuleId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
            const node = moduleGraph.getModuleById(virtualModuleId);
            return node && [node];
        },
    };
    async function resolveAbsoluteModuleId(id) {
        return ((await resolveBaseFilePath(id)) ||
            (await resolveBaseFilePath(path_1.default.join(rootDirPath, id))));
    }
    async function resolveBaseFilePath(baseFilePath) {
        for (const suffix of [
            "",
            ...jsExtensions,
            ...[...jsExtensions].map((extension) => `${path_1.default.sep}index${extension}`),
        ]) {
            const absoluteFilePath = `${baseFilePath}${suffix}`;
            const entry = await reader.read(absoluteFilePath);
            if (entry !== null) {
                return [absoluteFilePath, entry];
            }
        }
        return null;
    }
}
exports.virtualPlugin = virtualPlugin;
